class UnknownOperatorError(Exception):
    def __init__(self, operator, message="The operator is unknown!"):
        self.operator = operator
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"The operator {self.operator} is unknown!"


from typing import List
from pyqurl.core import Query, QueryPagination, QuerySort
from pyqurl.helpers import str2bool
from pyqurl.operations import *
from pony.orm.core import desc




class PonyORMHelper:

    @classmethod
    def apply_filters(cls, query, filters):
        if filters is None:
            return query

        for f in filters:
            if f.operation == EQUAL:
                query = query.filter(lambda x: getattr(x, f.prop) == f.value)

            elif f.operation == NOT_EQUAL:
                query = query.filter(lambda x: getattr(x, f.prop) != f.value)
                
            elif f.operation == CONTAINS:
                if isinstance(f.value, str):
                    query = query.filter(lambda x: f.value.lower() in getattr(x, f.prop).lower())
                else:
                    query = query.filter(lambda x: f.value in getattr(x, f.prop))

            elif f.operation == NOT_NULL:
                if str2bool(f.value):
                    query = query.filter(lambda x: getattr(x, f.prop) is not None)
                else:
                    query = query.filter(lambda x: getattr(x, f.prop) is None)

        return query    

    @classmethod
    def apply_pagination(cls, query, pagination: QueryPagination):
        if pagination is None:
            return query

        return query.limit(pagination.limit, offset=pagination.offset)

    @classmethod
    def apply_sorting(cls, query, sorts: List[QuerySort]):
        if sorts is None:
            return query

        for sort in sorts:
            query = query.order_by(lambda x: desc(getattr(x, sort.prop))) \
                if sort.descending else \
                    query.order_by(lambda x: getattr(x, sort.prop))

        return query

    @classmethod
    def apply_query(cls, pony_query, query: Query):
        if query is None:
            return pony_query

        filtered_query = cls.apply_filters(pony_query, query.filters)
        ordered_query = cls.apply_sorting(filtered_query, query.sort)
        paginated_query = cls.apply_pagination(ordered_query, query.pagination)
        return paginated_query


import functools
from quart import request

from pyqurl.core import create_query_from_dict

def use_query_filters():
    def wrapper(fn):
        @functools.wraps(fn)
        async def wrapped(*args, **kwargs):
            query = create_query_from_dict(request.args)
            return await fn(*args, **kwargs, query=query)
        return wrapped
    return wrapper



from dataclasses import dataclass


@dataclass
class FilterOperation:
    value: str    


CONTAINS = FilterOperation("ct")
GREATER_THAN = FilterOperation("gt")
GREATER_THAN_OR_EQUAL = FilterOperation("gte")
LESSER_THAN = FilterOperation("lt"  )
LESSER_THAN_OR_EQUAL = FilterOperation("lte")
EQUAL = FilterOperation("eq")
NOT_EQUAL = FilterOperation("neq")
IN = FilterOperation("in")
NOT_IN = FilterOperation("nin")
RANGE = FilterOperation("rng")
NOT_NULL = FilterOperation("nn")


OPERATIONS = [
    CONTAINS,
    GREATER_THAN,
    GREATER_THAN_OR_EQUAL,
    LESSER_THAN,
    LESSER_THAN_OR_EQUAL,
    EQUAL,
    NOT_EQUAL,
    IN,
    NOT_IN,
    RANGE,
    NOT_NULL,
]


OPERATIONS_BY_VALUE = { o.value: o for o in OPERATIONS }





from dataclasses import dataclass
import re
from typing import Any, Dict, List, Union
from urllib.parse import parse_qs
from datetime import datetime

from pyqurl.operations import *
from pyqurl.helpers import str2bool
from pyqurl.exceptions import UnknownOperatorError


@dataclass
class QueryFilter:
    prop: str
    operation: FilterOperation


@dataclass
class SingleValueQueryFilter(QueryFilter):
    value: Any


@dataclass
class ArrayValueQueryFilter(QueryFilter):
    value: List[Any]


@dataclass
class RangeValueQueryFilter(QueryFilter):
    start: Any
    end: Any


@dataclass
class QuerySort:
    prop: str
    descending: bool


@dataclass
class QueryPagination:
    limit: int
    offset: Union[int, str]


@dataclass
class Query:
    filters: List[QueryFilter]
    sort: List[QuerySort]
    pagination: QueryPagination


QUERY_KEY_REGEX = r"(\w*)[\[]?(\w*)?[\]]?"
QUERY_SORT_REGEX = r"([-+])?(\w*)"

def create_query_from_string(q: str) -> Query:
    d = parse_qs(q)
    return create_query_from_dict(d)


def create_query_from_dict(d: Dict) -> Query:
    args = {**d}

    offset = args.pop("offset", 0)
    try:
        offset = offset[0]
    except IndexError:
        pass

    try:
        offset = int(offset)
    except ValueError:
        pass

    limit = args.pop("limit", None)
    try:
        limit = limit[0]
    except IndexError:
        pass

    pagination = QueryPagination(int(limit), offset) if limit else None

    sort = create_sort_from_string(args.pop("sort", None))

    filters = create_filters_from_dict(args)

    return Query(filters, sort, pagination)


def parse_string_value(value: str) -> Any:
    if value is None or value.lower().strip() == "null":
        return None

    try:
        return datetime.fromisoformat(value)
    except ValueError:
        pass

    try:
        return int(value)
    except ValueError:
        pass

    try:
        return float(value)
    except ValueError:
        pass

    return value


def create_filters_from_dict(d: Dict):
    filters = []
    for key, value in d.items():
        if key is None or key.strip() == "":
            continue

        matches = re.search(QUERY_KEY_REGEX, key)
        prop, opvalue = matches.groups()

        if not opvalue:
            opvalue = EQUAL.value

        # Ensure value is a list
        if isinstance(value, List):
            value = ",".join(value).split(",")
        else:
            value = [value]

        # Try map every value to a number
        for i in range(len(value)):
            value[i] = parse_string_value(value[i])


        try:
            op = OPERATIONS_BY_VALUE[opvalue]
        except KeyError:
            raise UnknownOperatorError(opvalue)

        if op in [IN, NOT_IN]:
            f = ArrayValueQueryFilter(prop, op, value)

        elif op == NOT_NULL:
            f = SingleValueQueryFilter(prop, op, str2bool(value[0]))

        elif op == RANGE:
            start, end = value if value is not None else (None, None, )
            f = RangeValueQueryFilter(prop, op, start=start, end=end)

        else:
            f = SingleValueQueryFilter(prop, op, value[0])

        filters.append(f)

    return filters

def create_sort_from_string(value: Union[str, List]):
    if value is None:
        return None
    r = []

    if isinstance(value, List):
        value = ",".join(value).split(",")
    else:
        value = [value]

    for v in value:
        if not v:
            continue
        order, prop = re.search(QUERY_SORT_REGEX, v.strip()).groups()
        r.append(QuerySort(prop, order == "-"))
    return r








import pytest

from pyqurl import create_query_from_string
from pyqurl.exceptions import UnknownOperatorError
from pyqurl.operations import *


def test_single_explicit_equal_operator():
    query = create_query_from_string("name[eq]=gabriel")
    assert query.filters[0].prop == "name"
    assert query.filters[0].operation == EQUAL
    assert query.filters[0].value == "gabriel"


def test_single_implicit_equal_operator():
    query = create_query_from_string("name=gabriel")
    assert query.filters[0].prop == "name"
    assert query.filters[0].operation == EQUAL
    assert query.filters[0].value == "gabriel"


def test_multiple_equal_operators():
    query = create_query_from_string("name=john&surname=doe")

    assert query.filters[0].prop == "name"
    assert query.filters[0].operation == EQUAL
    assert query.filters[0].value == "john"

    assert query.filters[1].prop == "surname"
    assert query.filters[1].operation == EQUAL
    assert query.filters[1].value == "doe"

def test_query_without_value():
    query = create_query_from_string("name=")
    assert len(query.filters) == 0

def test_query_without_prop():
    query = create_query_from_string("=value")
    assert len(query.filters) == 0

def test_single_contains_operator():
    query = create_query_from_string("name[ct]=something")
    assert query.filters[0].prop == "name"
    assert query.filters[0].operation == CONTAINS
    assert query.filters[0].value == "something"


def test_single_notnull_operator_should_be_true():
    query = create_query_from_string("deleted[nn]=true")
    assert query.filters[0].prop == "deleted"
    assert query.filters[0].operation == NOT_NULL
    assert query.filters[0].value == True


def test_single_notnull_operator_should_be_false():
    query = create_query_from_string("deleted[nn]=false")
    assert query.filters[0].prop == "deleted"
    assert query.filters[0].operation == NOT_NULL
    assert query.filters[0].value == False


def test_unknown_operator_should_raise_exception():
    with pytest.raises(UnknownOperatorError):
        create_query_from_string("abc[aaaaa]=abc")


def test_single_in_operator():
    query = create_query_from_string("color[in]=red,green,blue")
    assert query.filters[0].prop == "color"
    assert query.filters[0].operation == IN
    assert query.filters[0].value == ["red", "green", "blue"]


def test_single_in_operator():
    query = create_query_from_string("year[rng]=1990,2010")
    assert query.filters[0].prop == "year"
    assert query.filters[0].operation == RANGE
    assert query.filters[0].start == 1990
    assert query.filters[0].end == 2010


def test_pagination():
    query = create_query_from_string("offset=20&limit=10")
    assert query.pagination is not None
    assert query.pagination.offset == 20
    assert query.pagination.limit == 10

def test_pagination_with_string_offset():
    query = create_query_from_string("offset=abc&limit=10")
    assert query.pagination is not None
    assert query.pagination.offset == "abc"
    assert query.pagination.limit == 10